function [coefs, dims] = WA_readCoefficients(fileName,intType, dataType)
% Reads Wave Atom coefficients from the file fileName
% coefs contains the coefficients and dims contains the dimensions
% intType should be the type corresponding to size_t c-type.
% Usually 'int32' for 32-bit system and 'int64' for 64-bit system.
% Currently WaveAtom package supports only dataType 'double'

%% Open output file
    [f,msg] = fopen(fileName,'r');
    if(isempty(msg)==0)
        error(msg)
        return
    end
    
%% Read dimension header
    [info,c] = fread(f,1,intType);
    if (c ~= 1)    
        fclose(f);
        error('Failed to read the number of matrix dimensions.')
    end 

    [dims,c] = fread(f,info,intType);
    if (c ~= info)    
        fclose(f);
        error('Failed to read the dimensions of the coefficients.')
    end
    
%% Read coefficients
    [coefs, c] = fread(f,dataType);    
    coefs = coefs(1:2:end) + coefs(2:2:end)*1i;
    if (c ~= 2*prod(dims))    
        fclose(f);
        error('Failed to read the correct number of coefficients.')
    end    

%% Close output file    
    fclose(f);
    return

end
