function [data] = WA_readDatainput(fileName,intType, dataType)
% Reads the data from the file specified by fileName and returns read data
% in data.
% intType should be the type corresponding to size_t c-type.
% Usually 'int32' for 32-bit system and 'int64' for 64-bit system.
% Currently WaveAtom package supports only dataType 'double' 

%% Open data file
    [f,msg] = fopen(fileName,'r');
    if(isempty(msg)==0)
        error(msg)
        return
    end

%% Write dimension header
    [info,c] = fread(f,1,intType);
    if (c ~= 1)    
        fclose(f);
        error('Failed to read the number of dimensions in data.')
    end 

    [dims,c] = fread(f,info,intType);
    if (c ~= info)    
        fclose(f);
        error('Failed to read the dimensions of the data.')
    end
    
%% Read data elements

    %Fix for 1D-case
    if (length(dims) == 1)
        dims = [dims(1);1];
    end
    
    %Read elements
    [data, c] = fread(f,dataType);     
    
    if (c ~= 2*prod(dims))    
        error('Failed to read the correct number of data elements.')
    end  
	data = data(1:2:end) + data(2:2:end)*1i;
    
    fclose(f);
    
    %Change row-order to column order
    if(length(dims)==2) %1D and 2D case 
        data = reshape(data,dims(end:-1:1).').';
    elseif(length(dims)==3) %3D case
        c = 0;
        temp = zeros(dims(1),dims(2),dims(3));
        for k = 1:dims(1)
            temp(k,:,:) = reshape(data((1:prod(dims(2:3)))+c),dims(3),dims(2)).';
            c = c + prod(dims(2:3));
        end               
        data = temp;
    elseif(length(dims)==4) %4D case
        c = 0;
        temp = zeros(dims(1),dims(2),dims(3),dims(4));
        for k = 1:dims(1)
            for l = 1:dims(2)
                temp(k,l,:,:) = reshape(data((1:prod(dims(3:4)))+c),dims(4),dims(3)).';
                c = c + prod(dims(3:4));
            end
        end               
        data = temp;    
    elseif(length(dims)==5) %5D case
        c = 0;
        temp = zeros(dims(1),dims(2),dims(3),dims(4),dims(5));
        for k = 1:dims(1)
            for l = 1:dims(2)
                for m = 1:dims(3)
                    temp(k,l,m,:,:) = reshape(data((1:prod(dims(4:5)))+c),dims(5),dims(4)).';
                    c = c + prod(dims(4:5));
                end
            end
        end               
        data = temp;    
    else %Higher dimensions are not supported at the moment
        error('Dimension is greater than 5. Not supported at the moment.')        
    end
    return

end
