function [] = WA_writeDatainput(fileName, mat, intType, dataType)
% Writes the real valued matrix mat into file fileName in row-major order
% intType should be the type corresponding to size_t c-type.
% Usually 'int32' for 32-bit system and 'int64' for 64-bit system.
% Currently WaveAtom package supports only dataType 'double' 

%% Open output file
    
    [f,msg] = fopen(fileName,'w');
    if(isempty(msg)==0)
        error(msg)        
    end
    
    
%% Write dimension header

    dimInfo = size(mat);
    
    if(dimInfo(2) == 1)
        dimInfo = dimInfo(1);        
    elseif(dimInfo(1) == 1)
        dimInfo = dimInfo(2);         
    end   
    
    c = fwrite(f,length(dimInfo),intType);
    if (c ~= 1)    
        fclose(f);
        error('Failed to write the number of dimensions.')
    end

    c = fwrite(f,dimInfo,intType);
    if (c ~= length(dimInfo))
        fclose(f);
        error('Failed to write the size of dimensions.')
    end
    
%% Write matrix mat in row-major order into output file
% Matlab uses column-order, but WaveAtom package expects the data to be in row-order.
    
    if(length(dimInfo)==1) %1D case
        temp = mat.';
        temp = temp(:).';
        temp = [real(temp);imag(temp)];
        temp = temp(:).';
        c = fwrite(f,temp,dataType);
    elseif(length(dimInfo)==2) %2D case
        temp = mat.';
        temp = temp(:).';
        temp = [real(temp);imag(temp)];
        temp = temp(:).';
        c = fwrite(f,temp,dataType);    
    elseif(length(dimInfo)==3) %3D case
        c = 0;
        for k = 1:dimInfo(1)
            temp = reshape(mat(k,:,:),dimInfo(2:3)).';
            temp = temp(:).';
            temp = [real(temp);imag(temp)];
            temp = temp(:).';
            c = c + fwrite(f,temp,dataType);
        end            
    elseif(length(dimInfo)==4) %4D case
        c = 0;
        for k = 1:dimInfo(1)
            for l = 1:dimInfo(2)
                temp = reshape(mat(k,l,:,:),dimInfo(3:4)).';
                temp = temp(:).';
                temp = [real(temp);imag(temp)];
                temp = temp(:).';
                c = c + fwrite(f,temp,dataType);
            end        
        end    
    elseif(length(dimInfo)==5) %5D case
        c = 0;
        for k = 1:dimInfo(1)
            for l = 1:dimInfo(2)
                for m = 1:dimInfo(3)
                    temp = reshape(mat(k,l,m,:,:),dimInfo(4:5)).';
                    temp = temp(:).';
                    temp = [real(temp);imag(temp)];
                    temp = temp(:).';
                    c = c + fwrite(f,temp,dataType);
                end
            end        
        end    
    else %Higher dimensions are not supported at the moment
      error('Dimension is greater than 5. Not supported at the moment.')
    end
    
    if (c ~= 2*numel(mat))
        fclose(f);
        error('Failed to write matrix elements.')                
    end
    
%% Close output file
    fclose(f);

end

