#include <stdio.h>
#include <time.h>
#include <limits.h>
#include <stdlib.h>
#include <complex.h>

unsigned int getseed();

int main(int argc, char *argv[])
{
  if(argc < 2){
    printf("Usage: WA_datagen [datafile] n_1 n_2 n_3 ..., where n_i is the number of elements in dimension i.\n");
    return 1;
  }

  size_t ndim = (size_t) (argc-2);


  FILE *fp = fopen(argv[1],"wb");
  if(fp == NULL){
    fprintf(stderr,"Failed to open file %s\n.",argv[1]);
    return 1;
  }

  if (fwrite(&ndim,sizeof(size_t),1,fp)!=1){
    fprintf(stderr,"Failed to write the number of dimensions.\n");
    fclose(fp);
    return 1;
  }

  size_t n = 1;
  for (int i = 2; i<argc; i++){
    size_t t = (size_t)abs(atol(argv[i]));
    n *= t;      
    if(fwrite(&t,sizeof(size_t),1,fp)!=1){
      fprintf(stderr,"Failed to write the size of dimension %d.\n",i);
      fclose(fp);
      return 1;
    }
  }

  if (n<1){
    fprintf(stderr,"Total number of elements is less than 1! (%zu)\n",n);
    fclose(fp);
    return 1;
  }

  printf("Total number of elements is %zu.\n",n);
  double* data = (double*)malloc(n*sizeof(double complex));
  if (data==NULL){
    fprintf(stderr,"Failed to allocate enough memory!\n");
    fclose(fp);
    return 1;
  }

  srand (getseed());

  for(size_t d = 0; d<2*n;++d){
    data[d] = rand()/(RAND_MAX + 1.0);
  }

  size_t r = fwrite(data,sizeof(double complex),n,fp);
  printf("Wrote %zu elements into file %s.\n",r,argv[1]); 
  
  fclose(fp);
  
}

unsigned int getseed()
{
  time_t t = time(0);
  unsigned char *p = (unsigned char*)&t;
  unsigned int seed = 0;
  size_t i;
  
  for (i=0;i<sizeof(t);++i)
     seed = seed*(UCHAR_MAX+2U)+p[i];
  
  return seed;
}

