#include "WA_iwa.h"

bool iwa1(inputData *id, double complex* coefficients, bool uniformpartition,unsigned fftw_flags)
{
  /*
    Check that input data has correct number of dimensions.
   */
  
  printf("Calculating 1D inverse wave atom transform...\n");

  if(id->numberOfDimensions != 1){
    fprintf(stderr,"Number of dimensions is not one! Instead, it is (%zu).\n", id->numberOfDimensions);
    return false;
  }

  /* 
     Reserve enough memory for different arrays, perform other initializations, and generate frequency partition.
   */
  
  size_t sizeOfSmallestDimension = id->sizeOfDimension[0];
  
  frequencyPartition freq_par;
  bool (*partitiofunc)(const size_t, frequencyPartition *);
  if (uniformpartition){
    partitiofunc = UniformFrequencyPartition;
  } else {
    partitiofunc = OrthogonalFrequencyPartition;
  }
  
  if (!partitiofunc(sizeOfSmallestDimension, &freq_par)){
    return false;
  }
  PrintFrequencyPartition(&freq_par);  
  
  size_t pos;  

  fftw_complex *outputData = (fftw_complex*)id->data;
  size_t nCoefInCal = 1;

  fftw_complex* wacoef = (fftw_complex*) malloc((2<<freq_par.maxJ)*sizeof(fftw_complex));      
  if (wacoef == NULL){
    fprintf(stderr,"Error when reserving temporary memory.");
    FreeFrequencyPartition(&freq_par);
    free(wacoef);
    return false;
  }  
  memset(outputData,0,sizeof(double complex)*id->numberOfElements);
  
  fftw_complex* tempmemorystart = coefficients;
  fftw_complex* tempmemory = coefficients;

  /*
    Loop over scale j
	Perform inverse FFT to each coefficient set (j, m).
	Normalize each coefficient set (j, m).
        Loop over frequency partition m
            Calculate required Mayer's Wavelet values.
            Calculate contributions to output from coefficient set (j, m).
	End
    End
    Return output
   */

  for (size_t j=0; j<=freq_par.maxJ;++j){
    printf("Scale j=%zu...\n",j);
    nCoefInCal <<= 1;

    double b = 1<<j;    
    long long int dx1;
    long long int dx2;
    long long int p1;
    long long int p2;
    double dx = 1/b; 
    double fm;    
    double d,t;
    double complex c;

    int tempn = (int) nCoefInCal;    
    fftw_plan p = fftw_plan_many_dft(1,&tempn,freq_par.lengthOfM[j]-freq_par.startOfM[j],tempmemorystart,&tempn,1, nCoefInCal, tempmemorystart,&tempn,1,nCoefInCal,FFTW_FORWARD,fftw_flags);   
    fftw_execute(p);
    fftw_destroy_plan(p);
    
    double normalization = 1./sqrt(nCoefInCal);    
    for(size_t i = 0; i<(freq_par.lengthOfM[j]-freq_par.startOfM[j])*nCoefInCal;++i){
      (*tempmemorystart) *= normalization;
      tempmemorystart++;
    }   
    
    for(size_t m1 = freq_par.startOfM[j];m1 < freq_par.lengthOfM[j];++m1){      
      pos = nCoefInCal-1;

      if(m1%2==0){
	fm = (m1-2/3.)*b;
	dx1 = ceil(fm);

	d = dx1*dx-m1;
	t = M_PI_2*(d-0.5);

	for(size_t i = 0; i < nCoefInCal; ++i){
	  c = cos(t)-sin(t)*I;
	  if(4/3. < d || d < -2/3.) wacoef[i] = 0;
	  else if (2/3. < d) wacoef[i] = c*sin(M_PI_4*(1+cos(M_PI*(1-1.5*d))));
	  else wacoef[i] = c*cos(M_PI_4*(1-cos(M_PI*(0.5-0.75*d))));
	  d += dx;  
	  t += M_PI_2*dx;
	}	

      } else {	
	fm = (m1-1/3.)*b;
	dx1 = ceil(fm);

	d = dx1*dx-m1;
	t = M_PI_2*(d-0.5);
	for(size_t i = 0; i < nCoefInCal; ++i){
	  c = cos(t)-sin(t)*I;
	  if (d < -1/3. || 5/3. < d) wacoef[i] = 0;
	  if (d < 1/3.) wacoef[i] = c*sin(M_PI_4*(1+cos(M_PI*(1.5*d-0.5))));
	  else wacoef[i] = c*cos(M_PI_4*(1-cos(M_PI*(0.75*d-0.25))));
	  d += dx;  
	  t += M_PI_2*dx;
	}		

      }

      dx2 = dx1+nCoefInCal-1;
      
      for (size_t y=0; y<nCoefInCal;++y){	
	p1 = dx1%nCoefInCal;
	p2 = dx1%sizeOfSmallestDimension;
	
	if(p1<0) p1+=nCoefInCal;
	if(p2<0) p2+=sizeOfSmallestDimension;
	
	outputData[p2] += wacoef[y]*tempmemory[p1];

	p1 = (-dx2)%nCoefInCal;
	p2 = (-dx2)%sizeOfSmallestDimension;	
	
	if(p1<0) p1+=nCoefInCal;
	if(p2<0) p2+=sizeOfSmallestDimension;
	
	outputData[p2] += conj(wacoef[pos--])*tempmemory[p1];

	dx1++;
	dx2--;
      }
      tempmemory += nCoefInCal;
    }
  }  
  free(wacoef);
  FreeFrequencyPartition(&freq_par);
  return true;
}

