#include "WA_cuda.h"
#include "WA_cuda_kernels.h"
#include <thrust/version.h>
#include <thrust/device_ptr.h>

struct constmultiply_functor:public thrust::unary_function<cuDoubleComplex,cuDoubleComplex>
{
  const cuDoubleComplex a;

  constmultiply_functor(double _a) : a(make_cuDoubleComplex(_a,0)) {}
  constmultiply_functor(cuDoubleComplex _a) : a(_a) {}
  __host__ __device__
  cuDoubleComplex operator()(const cuDoubleComplex& x) const { 
    return cuCmul(x,a);    
  }
};


void thrustVersion(void)
{
  int major = THRUST_MAJOR_VERSION;
  int minor = THRUST_MINOR_VERSION;

  printf("Thrust v%d.%d\n", major, minor);

}

void thrustMultiplyOnDevice(cuDoubleComplex* data, size_t numElems, double m)
{
/*
  cudaEvent_t start, stop;
  float elapsedTime;
  CudaSafeCall(cudaEventCreate(&start));
  CudaSafeCall(cudaEventCreate(&stop));    
  printf("Thrust multiply on device...\n");
  CudaSafeCall(cudaEventRecord(start,0));
*/
  thrust::device_ptr<cuDoubleComplex> dev_ptr(data);
  thrust::transform(dev_ptr,dev_ptr+(int)numElems,dev_ptr,constmultiply_functor(m));
/*
  CudaSafeCall(cudaEventRecord(stop,0));
  CudaSafeCall(cudaEventSynchronize(stop));
  CudaSafeCall(cudaEventElapsedTime(&elapsedTime,start,stop));
  printf("Thrust multiply on device took %3.1f ms.\n", elapsedTime);
*/

}

#define WA_GPU_BLOCKSIZE_2D 16
#define WA_GPU_BLOCKSIZE_3D 8
void kernel2D_gpu_wrapper(int i1, int i2,	      
			  int sm, int lm, 
			  int startBlock, int numberOfBlocks,
			  int actuallyCalcBlocksAlready,
			  int dime1, int dime2,
			  int rnc1, int rnc2,
			  int nelemsj, 
			  cuDoubleComplex *data,
			  cuDoubleComplex *tempmemory,
			  cuDoubleComplex *dataws1, 
			  cuDoubleComplex *dataws2,
			  int *wacoefStart
			  )
{
  dim3 grid(numberOfBlocks,1,1);
  dim3 block(WA_GPU_BLOCKSIZE_2D,WA_GPU_BLOCKSIZE_2D,1);
  kernel2D_gpu<<<grid, block>>>(i1,i2,sm,lm,startBlock,
				actuallyCalcBlocksAlready,
				dime1,dime2,rnc1,rnc2,
				nelemsj, data,
				tempmemory, 
				dataws1, dataws2,
				wacoefStart);
  //printf("Finished 2D-kernel run with %d blocks\n",numberOfBlocks);
}

void inverse_kernel2D_gpu_wrapper(int i1, int i2,	      
				  int p1, int p2,
				  int sm, int lm, 
				  int startBlock, int numberOfBlocks,
				  int actuallyCalcBlocksAlready,
				  int dime1, int dime2,
				  int rnc1, int rnc2,
				  int nelemsj, 
				  cuDoubleComplex *data,
				  cuDoubleComplex *tempmemory,
				  cuDoubleComplex *dataws1, 
				  cuDoubleComplex *dataws2,
				  int *wacoefStart
				  )
{
  dim3 grid(numberOfBlocks,1,1);
  dim3 block(WA_GPU_BLOCKSIZE_2D,WA_GPU_BLOCKSIZE_2D,1);
  inverse_kernel2D_gpu<<<grid, block>>>(i1,i2,p1,p2,sm,lm,startBlock,
				actuallyCalcBlocksAlready,
				dime1,dime2,rnc1,rnc2,
				nelemsj, data,
				tempmemory, 
				dataws1, dataws2,
				wacoefStart);
  //printf("Finished inverse 2D-kernel run with %d blocks\n",numberOfBlocks);
}


void kernel3D_gpu_wrapper(int i1, int i2, int i3,
			  int sm, int lm, 
			  int startBlock, int numberOfBlocks,
			  int actuallyCalcBlocksAlready,
			  int dime1, int dime2, int dime3,
			  int rnc1, int rnc2, int rnc3,
			  int nelemsj, 
			  cuDoubleComplex *data,
			  cuDoubleComplex *tempmemory,
			  cuDoubleComplex *dataws1, 
			  cuDoubleComplex *dataws2,
			  cuDoubleComplex *dataws3,
			  int *wacoefStart
			  )
{
  dim3 grid(numberOfBlocks,1,1);
  dim3 block(WA_GPU_BLOCKSIZE_3D,WA_GPU_BLOCKSIZE_3D,WA_GPU_BLOCKSIZE_3D);
  kernel3D_gpu<<<grid, block>>>(i1,i2,i3,sm,lm,startBlock,
				actuallyCalcBlocksAlready,
				dime1,dime2,dime3,rnc1,rnc2,rnc3,
				nelemsj, data,
				tempmemory, 
				dataws1, dataws2,dataws3,
				wacoefStart);

//  printf("Finished 3D-kernel run with %d blocks\n",numberOfBlocks);
}

void inverse_kernel3D_gpu_wrapper(int i1, int i2, int i3,
				  int p1, int p2, int p3,
				  int sm, int lm, 
				  int startBlock, int numberOfBlocks,
				  int actuallyCalcBlocksAlready,
				  int dime1, int dime2, int dime3,
				  int rnc1, int rnc2, int rnc3,
				  int nelemsj, 
				  cuDoubleComplex *data,
				  cuDoubleComplex *tempmemory,
				  cuDoubleComplex *dataws1, 
				  cuDoubleComplex *dataws2,
				  cuDoubleComplex *dataws3,
				  int *wacoefStart
				  )
{
  dim3 grid(numberOfBlocks,1,1);
  dim3 block(WA_GPU_BLOCKSIZE_3D,WA_GPU_BLOCKSIZE_3D,WA_GPU_BLOCKSIZE_3D);
  inverse_kernel3D_gpu<<<grid, block>>>(i1,i2,i3,
					p1,p2,p3,
					sm,lm,startBlock,
					actuallyCalcBlocksAlready,
					dime1,dime2,dime3,
					rnc1,rnc2,rnc3,
					nelemsj, data,
					tempmemory, 
					dataws1, dataws2,dataws3,
					wacoefStart);

  //printf("Finished inverse 3D-kernel run with %d blocks\n",numberOfBlocks);
}



__global__ void kernel2D_gpu(int i1, int i2,	      
			     int sm, int lm,
			     int mstart,int actuallyCalcBlocksAlready,
			     int dime1, int dime2,
			     int rnc1, int rnc2,
			     int nelemsj, 
			     cuDoubleComplex *data,
			     cuDoubleComplex *tempmemory,
			     cuDoubleComplex *dataws1, 
			     cuDoubleComplex *dataws2,
			     int *wacoefStart
			     )
{
  
  int m1 = (blockIdx.x+mstart)/lm;
  int m2 = (blockIdx.x+mstart)%lm;  

  if(m1>=sm || m2>=sm){

    int s1 = threadIdx.x;
    int s2 = threadIdx.y;

    size_t ks1 = m1*rnc1;
    size_t ks2 = lm*rnc1 + m2*rnc2;
  
    int temp;

    int sposModJ0;
    int sposModJ0Start;
    int sposModDim0;
    int sposModDim0Start;

    int sposModJ1;
    int sposModJ1Start;
    int sposModDim1;
    int sposModDim1Start;

    if (m1 >= sm){
      tempmemory += ((lm-sm)*sm+(m1-sm)*lm+m2)*nelemsj;
    } else {
      tempmemory += ((lm-sm)*m1+m2-sm)*nelemsj;
    }
    
    tempmemory -= actuallyCalcBlocksAlready*nelemsj;

    if(i1 == 0){
      temp = wacoefStart[m1] + s1;
    } else {
      temp = -(wacoefStart[m1]+rnc1-1) + s1;
    }    
    sposModJ0Start = temp&(rnc1-1);
    sposModDim0Start = temp&(dime1-1);
    if (sposModJ0Start < 0)
      sposModJ0Start += rnc1;
    if (sposModDim0Start< 0)
      sposModDim0Start += dime1;


    if(i2 == 0){
      temp = wacoefStart[lm+m2] + s2;
    } else {
      temp = -(wacoefStart[lm+m2]+rnc2-1) + s2;
    }
    sposModJ1Start = temp&(rnc2-1);
    sposModDim1Start = temp&(dime2-1);
    if (sposModJ1Start < 0)
      sposModJ1Start += rnc2;
    if (sposModDim1Start< 0)
      sposModDim1Start += dime2;

    sposModJ0 = sposModJ0Start;
    sposModDim0 = sposModDim0Start;
    while(s1<rnc1){
      sposModJ1 = sposModJ1Start;
      sposModDim1 = sposModDim1Start;	
      while(s2<rnc2){
	size_t sijoituspossi =
	  sposModJ0*rnc2+sposModJ1;
	size_t datapossi = 
	  sposModDim0*dime2+sposModDim1;
	
	tempmemory[sijoituspossi] = 
	  cuCadd(tempmemory[sijoituspossi],
		 cuCmul(cuCmul(dataws1[ks1+s1],dataws2[ks2+s2]),data[datapossi]));

	sposModJ1 = (sposModJ1+WA_GPU_BLOCKSIZE_2D)&(rnc2-1);
	sposModDim1 = (sposModDim1+WA_GPU_BLOCKSIZE_2D)&(dime2-1);	  
	s2 += WA_GPU_BLOCKSIZE_2D;
      }
      sposModJ0 = (sposModJ0+WA_GPU_BLOCKSIZE_2D)&(rnc1-1);
      sposModDim0 = (sposModDim0+WA_GPU_BLOCKSIZE_2D)&(dime1-1);	  

      s1 += WA_GPU_BLOCKSIZE_2D;
      s2 = threadIdx.y;
    }
  }    
}

__global__ void inverse_kernel2D_gpu(int i1, int i2,	      
				     int p1, int p2,
				     int sm, int lm,
				     int mstart,int actuallyCalcBlocksAlready,
				     int dime1, int dime2,
				     int rnc1, int rnc2,
				     int nelemsj, 
				     cuDoubleComplex *data,
				     cuDoubleComplex *tempmemory,
				     cuDoubleComplex *dataws1, 
				     cuDoubleComplex *dataws2,
				     int *wacoefStart
				     )
{  
  int m1 = 2*((blockIdx.x+mstart)/(lm>>1))+p1;
  int m2 = 2*((blockIdx.x+mstart)%(lm>>1))+p2;  

  if(m1<lm && m2<lm)
  if(m1>=sm || m2>=sm){

    int s1 = threadIdx.x;
    int s2 = threadIdx.y;

    size_t ks1 = m1*rnc1;
    size_t ks2 = lm*rnc1 + m2*rnc2;
  
    int temp;

    int sposModJ0;
    int sposModJ0Start;
    int sposModDim0;
    int sposModDim0Start;

    int sposModJ1;
    int sposModJ1Start;
    int sposModDim1;
    int sposModDim1Start;

    if (m1 >= sm){
      data += ((lm-sm)*sm+(m1-sm)*lm+m2)*nelemsj;
    } else {
      data += ((lm-sm)*m1+m2-sm)*nelemsj;
    }
    
    data -= actuallyCalcBlocksAlready*nelemsj;

    if(i1 == 0){
      temp = wacoefStart[m1] + s1;
    } else {
      temp = -(wacoefStart[m1]+rnc1-1) + s1;
    }    
    sposModJ0Start = temp&(rnc1-1);
    sposModDim0Start = temp&(dime1-1);
    if (sposModJ0Start < 0)
      sposModJ0Start += rnc1;
    if (sposModDim0Start< 0)
      sposModDim0Start += dime1;


    if(i2 == 0){
      temp = wacoefStart[lm+m2] + s2;
    } else {
      temp = -(wacoefStart[lm+m2]+rnc2-1) + s2;
    }
    sposModJ1Start = temp&(rnc2-1);
    sposModDim1Start = temp&(dime2-1);
    if (sposModJ1Start < 0)
      sposModJ1Start += rnc2;
    if (sposModDim1Start< 0)
      sposModDim1Start += dime2;

    sposModJ0 = sposModJ0Start;
    sposModDim0 = sposModDim0Start;
    while(s1<rnc1){
      sposModJ1 = sposModJ1Start;
      sposModDim1 = sposModDim1Start;	
      while(s2<rnc2){
	size_t sijoituspossi =
	  sposModJ0*rnc2+sposModJ1;
	size_t datapossi = 
	  sposModDim0*dime2+sposModDim1;
	
	tempmemory[datapossi] = 
	  cuCadd(tempmemory[datapossi],
		 cuCmul(cuCmul(dataws1[ks1+s1],dataws2[ks2+s2]),data[sijoituspossi]));

	sposModJ1 = (sposModJ1+WA_GPU_BLOCKSIZE_2D)&(rnc2-1);
	sposModDim1 = (sposModDim1+WA_GPU_BLOCKSIZE_2D)&(dime2-1);	  
	s2 += WA_GPU_BLOCKSIZE_2D;
      }
      sposModJ0 = (sposModJ0+WA_GPU_BLOCKSIZE_2D)&(rnc1-1);
      sposModDim0 = (sposModDim0+WA_GPU_BLOCKSIZE_2D)&(dime1-1);	  
 
      s1 += WA_GPU_BLOCKSIZE_2D;
      s2 = threadIdx.y;      
    }
  }    
}


__global__ void kernel3D_gpu(int i1, int i2, int i3,
			     int sm, int lm,
			     int mstart,int actuallyCalcBlocksAlready,
			     int dime1, int dime2, int dime3,
			     int rnc1, int rnc2, int rnc3,
			     int nelemsj, 
			     cuDoubleComplex *data,
			     cuDoubleComplex *tempmemory,
			     cuDoubleComplex *dataws1, 
			     cuDoubleComplex *dataws2,
			     cuDoubleComplex *dataws3,
			     int *wacoefStart
			     )
{
  int m3 = (blockIdx.x+mstart)%lm;
  int m2 = ((blockIdx.x+mstart-m3)/lm)%lm;
  int m1 = (blockIdx.x+mstart-m3-m2*lm)/(lm*lm);

  if(m1>=sm || m2>=sm || m3>=sm){    
    if(m1<sm){
      tempmemory += (lm*lm-sm*sm)*m1*nelemsj;
      if(m2<sm){
	tempmemory += (m2*(lm-sm)+m3-sm)*nelemsj; 
      } else {
	tempmemory += ((lm-sm)*sm + (m2-sm)*lm+m3)*nelemsj;
      }
    } else {
      tempmemory += ((lm*lm-sm*sm)*sm + ((m1-sm)*lm+m2)*lm+m3)*nelemsj;
    }

    tempmemory -= actuallyCalcBlocksAlready*nelemsj;
 
    int s1 = threadIdx.x;
    int s2 = threadIdx.y;
    int s3 = threadIdx.z;    
  
    int ks1 = m1*rnc1;
    int ks2 = lm*rnc1 + m2*rnc2;
    int ks3 = lm*(rnc1 + rnc2) + m3*rnc3;

    int temp;

    int sposModJ0;
    int sposModJ0Start;
    int sposModDim0;
    int sposModDim0Start;

    int sposModJ1;
    int sposModJ1Start;
    int sposModDim1;
    int sposModDim1Start;

    int sposModJ2;
    int sposModJ2Start;
    int sposModDim2;
    int sposModDim2Start;    

    if(i1 == 0){
      temp = wacoefStart[m1] + s1;
    } else {
      temp = -(wacoefStart[m1]+rnc1-1) + s1;
    }    
    sposModJ0Start = temp&(rnc1-1);
    sposModDim0Start = temp&(dime1-1);
    if (sposModJ0Start < 0)
      sposModJ0Start += rnc1;
    if (sposModDim0Start< 0)
      sposModDim0Start += dime1;

    if(i2 == 0){
      temp = wacoefStart[lm+m2] + s2;
    } else {
      temp = -(wacoefStart[lm+m2]+rnc2-1) + s2;
    }
    sposModJ1Start = temp&(rnc2-1);
    sposModDim1Start = temp&(dime2-1);   
    if (sposModJ1Start < 0)
      sposModJ1Start += rnc2;
    if (sposModDim1Start < 0)
      sposModDim1Start += dime2;

    if(i3 == 0){
      temp = wacoefStart[2*lm+m3] + s3;
    } else {
      temp = -(wacoefStart[2*lm+m3]+rnc3-1) + s3;
    }
    sposModJ2Start = temp&(rnc3-1);
    sposModDim2Start = temp&(dime3-1);	  

    if (sposModJ2Start < 0)
      sposModJ2Start += rnc3;
    if (sposModDim2Start < 0)
      sposModDim2Start += dime3;   

    sposModJ0 = sposModJ0Start;
    sposModDim0 = sposModDim0Start;
    while(s1<rnc1){
      sposModJ1 = sposModJ1Start;
      sposModDim1 = sposModDim1Start;	
      while(s2<rnc2){
	  sposModJ2 = sposModJ2Start;
	  sposModDim2 = sposModDim2Start;	
	while(s3<rnc3){	  

	  cuDoubleComplex kerroin = 
	    cuCmul(dataws1[ks1+s1],
		   cuCmul(dataws2[ks2+s2], 
			  dataws3[ks3+s3]));
	  
	  size_t sijoituspossi =
	    (sposModJ0*rnc2+
	     sposModJ1)*rnc3+
	    sposModJ2;
	  
	  size_t datapossi = 
	    (sposModDim0*dime2+
	     sposModDim1)*dime3+
	    sposModDim2;
	  
	  tempmemory[sijoituspossi] = 
	    cuCadd(tempmemory[sijoituspossi],
		   cuCmul(kerroin,data[datapossi]));
	  
	  sposModJ2 = (sposModJ2+WA_GPU_BLOCKSIZE_3D)&(rnc3-1);
	  sposModDim2 = (sposModDim2+WA_GPU_BLOCKSIZE_3D)&(dime3-1);	  
	  
	  s3 += WA_GPU_BLOCKSIZE_3D;
	}

	sposModJ1 = (sposModJ1+WA_GPU_BLOCKSIZE_3D)&(rnc2-1);
	sposModDim1 = (sposModDim1+WA_GPU_BLOCKSIZE_3D)&(dime2-1);	  
	s2 += WA_GPU_BLOCKSIZE_3D;
	s3 = threadIdx.z;
      }

      sposModJ0 = (sposModJ0+WA_GPU_BLOCKSIZE_3D)&(rnc1-1);
      sposModDim0 = (sposModDim0+WA_GPU_BLOCKSIZE_3D)&(dime1-1);	  
      s1 += WA_GPU_BLOCKSIZE_3D;
      s2 = threadIdx.y;
    }
  }    
}

__global__ void inverse_kernel3D_gpu(int i1, int i2, int i3,
				     int p1, int p2, int p3,
				     int sm, int lm,
				     int mstart,int actuallyCalcBlocksAlready,
				     int dime1, int dime2, int dime3,
				     int rnc1, int rnc2, int rnc3,
				     int nelemsj, 
				     cuDoubleComplex *data,
				     cuDoubleComplex *tempmemory,
				     cuDoubleComplex *dataws1, 
				     cuDoubleComplex *dataws2,
				     cuDoubleComplex *dataws3,
				     int *wacoefStart
				     )
{
  int hg = lm>>1;
  int m3 = (blockIdx.x+mstart)%hg;
  int m2 = ((blockIdx.x+mstart-m3)/hg)%hg;
  int m1 = (blockIdx.x+mstart-m3-m2*hg)/(hg*hg);

  m1 = 2*m1 + p1;
  m2 = 2*m2 + p2;
  m3 = 2*m3 + p3;

  if(m1<lm && m2<lm && m3<lm)
  if(m1>=sm || m2>=sm || m3>=sm){    
    if(m1<sm){
      data += (lm*lm-sm*sm)*m1*nelemsj;
      if(m2<sm){
	data += (m2*(lm-sm)+m3-sm)*nelemsj; 
      } else {
	data += ((lm-sm)*sm + (m2-sm)*lm+m3)*nelemsj;
      }
    } else {
      data += ((lm*lm-sm*sm)*sm + ((m1-sm)*lm+m2)*lm+m3)*nelemsj;
    }

    data -= actuallyCalcBlocksAlready*nelemsj;

    int s1 = threadIdx.x;
    int s2 = threadIdx.y;
    int s3 = threadIdx.z;    
  
    int ks1 = m1*rnc1;
    int ks2 = lm*rnc1 + m2*rnc2;
    int ks3 = lm*(rnc1 + rnc2) + m3*rnc3;

    int temp;

    int sposModJ0;
    int sposModJ0Start;
    int sposModDim0;
    int sposModDim0Start;

    int sposModJ1;
    int sposModJ1Start;
    int sposModDim1;
    int sposModDim1Start;

    int sposModJ2;
    int sposModJ2Start;
    int sposModDim2;
    int sposModDim2Start;    

    if(i1 == 0){
      temp = wacoefStart[m1] + s1;
    } else {
      temp = -(wacoefStart[m1]+rnc1-1) + s1;
    }    
    sposModJ0Start = temp&(rnc1-1);
    sposModDim0Start = temp&(dime1-1);
    if (sposModJ0Start < 0)
      sposModJ0Start += rnc1;
    if (sposModDim0Start< 0)
      sposModDim0Start += dime1;

    if(i2 == 0){
      temp = wacoefStart[lm+m2] + s2;
    } else {
      temp = -(wacoefStart[lm+m2]+rnc2-1) + s2;
    }
    sposModJ1Start = temp&(rnc2-1);
    sposModDim1Start = temp&(dime2-1);   
    if (sposModJ1Start < 0)
      sposModJ1Start += rnc2;
    if (sposModDim1Start < 0)
      sposModDim1Start += dime2;

    if(i3 == 0){
      temp = wacoefStart[2*lm+m3] + s3;
    } else {
      temp = -(wacoefStart[2*lm+m3]+rnc3-1) + s3;
    }
    sposModJ2Start = temp&(rnc3-1);
    sposModDim2Start = temp&(dime3-1);	  

    if (sposModJ2Start < 0)
      sposModJ2Start += rnc3;
    if (sposModDim2Start < 0)
      sposModDim2Start += dime3;   

    sposModJ0 = sposModJ0Start;
    sposModDim0 = sposModDim0Start;
    while(s1<rnc1){
      sposModJ1 = sposModJ1Start;
      sposModDim1 = sposModDim1Start;	
      while(s2<rnc2){
	  sposModJ2 = sposModJ2Start;
	  sposModDim2 = sposModDim2Start;	
	while(s3<rnc3){	  

	  cuDoubleComplex kerroin = 
	    cuCmul(dataws1[ks1+s1],
		   cuCmul(dataws2[ks2+s2], 
			  dataws3[ks3+s3]));
	  
	  size_t sijoituspossi =
	    (sposModJ0*rnc2+
	     sposModJ1)*rnc3+
	    sposModJ2;
	  
	  size_t datapossi = 
	    (sposModDim0*dime2+
	     sposModDim1)*dime3+
	    sposModDim2;

	  tempmemory[datapossi] = 
	    cuCadd(tempmemory[datapossi],
		   cuCmul(kerroin,data[sijoituspossi]));
	  
	  sposModJ2 = (sposModJ2+WA_GPU_BLOCKSIZE_3D)&(rnc3-1);
	  sposModDim2 = (sposModDim2+WA_GPU_BLOCKSIZE_3D)&(dime3-1);	  
	  
	  s3 += WA_GPU_BLOCKSIZE_3D;
	}

	sposModJ1 = (sposModJ1+WA_GPU_BLOCKSIZE_3D)&(rnc2-1);
	sposModDim1 = (sposModDim1+WA_GPU_BLOCKSIZE_3D)&(dime2-1);	  
	s2 += WA_GPU_BLOCKSIZE_3D;
	s3 = threadIdx.z;
      }

      sposModJ0 = (sposModJ0+WA_GPU_BLOCKSIZE_3D)&(rnc1-1);
      sposModDim0 = (sposModDim0+WA_GPU_BLOCKSIZE_3D)&(dime1-1);	  
      s1 += WA_GPU_BLOCKSIZE_3D;
      s2 = threadIdx.y;
    }
  }    
}





// General GPU Device CUDA Initialization
int gpuDeviceInit(int devID)
{
    int deviceCount;
    CudaSafeCall(cudaGetDeviceCount(&deviceCount));

    if (deviceCount == 0)
    {
        fprintf(stderr, "gpuDeviceInit() CUDA error: no devices supporting CUDA.\n");
        exit(-1);
    }

    if (devID < 0)
       devID = 0;
        
    if (devID > deviceCount-1)
    {
        fprintf(stderr, "\n");
        fprintf(stderr, ">> %d CUDA capable GPU device(s) detected. <<\n", deviceCount);
        fprintf(stderr, ">> gpuDeviceInit (-device=%d) is not a valid GPU device. <<\n", devID);
        fprintf(stderr, "\n");
        return -devID;
    }

    cudaDeviceProp deviceProp;
    CudaSafeCall( cudaGetDeviceProperties(&deviceProp, devID) );

    if (deviceProp.major < 1)
    {
        fprintf(stderr, "gpuDeviceInit(): GPU device does not support CUDA.\n");
        exit(-1);                                                  
     }
    
    CudaSafeCall( cudaSetDevice(devID) );
    printf("gpuDeviceInit() CUDA Device [%d]: \"%s\"\n", devID, deviceProp.name);

    return devID;
}

int findCudaDevice(void)
{
  int num_devices, device;
  int max_multiprocessors = 0, max_device = 0;

  CudaSafeCall(cudaGetDeviceCount(&num_devices));
  for (device = 0; device < num_devices; device++) {
    cudaDeviceProp properties;
    CudaSafeCall(cudaGetDeviceProperties(&properties, device));
    if (max_multiprocessors < properties.multiProcessorCount) {
      max_multiprocessors = properties.multiProcessorCount;
      max_device = device;
    }
  }
  
  gpuDeviceInit( max_device );
  return max_device;
}
