#ifndef WA_CUDA_INCLUDED
#define WA_CUDA_INCLUDED

#include <stdio.h>
#include <cuda.h>
#include <cufft.h>
#include <cuComplex.h>
#include <cuda_runtime_api.h>

#include "WA_cuda_errorhandling.h"

#ifdef __cplusplus
extern "C" {
#endif

  /*
    Functions to init device
   */
  int gpuDeviceInit(int devID);
  int findCudaDevice(void);

  /*
    Functions to perform normalizations using thrust
   */
  void thrustVersion(void);
  void thrustMultiplyOnDevice(cuDoubleComplex* data, size_t numElems, double m);

  /*
    Wrappers to CUDA kernels used in 2D and 3D forward and inverse Wave Atom transfers.
   */
  void kernel2D_gpu_wrapper(int i1, int i2,	      
			    int sm, int lm, 
			    int startBlock, int numberOfBlocks,
			    int actuallyCalcBlocksAlready,
			    int dime1, int dime2,
			    int rnc1, int rnc2,
			    int nelemsj, 
			    cuDoubleComplex *data,
			    cuDoubleComplex *tempmemory,
			    cuDoubleComplex *dataws1, 
			    cuDoubleComplex *dataws2,
			    int *wacoefStart
			  );

  void inverse_kernel2D_gpu_wrapper(int i1, int i2,	      
				    int p1, int p2,
				    int sm, int lm, 
				    int startBlock, int numberOfBlocks,
				    int actuallyCalcBlocksAlready,
				    int dime1, int dime2,
				    int rnc1, int rnc2,
				    int nelemsj, 
				    cuDoubleComplex *data,
				    cuDoubleComplex *tempmemory,
				    cuDoubleComplex *dataws1, 
				    cuDoubleComplex *dataws2,
				    int *wacoefStart
				    );

  
  void kernel3D_gpu_wrapper(int i1, int i2, int i3,	      
			    int sm, int lm, 
			    int startBlock, int numberOfBlocks,
			    int actuallyCalcBlocksAlready,
			    int dime1, int dime2, int dime3,
			    int rnc1, int rnc2, int rnc3,
			    int nelemsj, 
			    cuDoubleComplex *data,
			    cuDoubleComplex *tempmemory,
			    cuDoubleComplex *dataws1, 
			    cuDoubleComplex *dataws2,
			    cuDoubleComplex *dataws3,
			    int *wacoefStart
			    );

  void inverse_kernel3D_gpu_wrapper(int i1, int i2, int i3,	      
				    int p1, int p2, int p3,	      
				    int sm, int lm, 
				    int startBlock, int numberOfBlocks,
				    int actuallyCalcBlocksAlready,
				    int dime1, int dime2, int dime3,
				    int rnc1, int rnc2, int rnc3,
				    int nelemsj, 
				    cuDoubleComplex *data,
				    cuDoubleComplex *tempmemory,
				    cuDoubleComplex *dataws1, 
				    cuDoubleComplex *dataws2,
				    cuDoubleComplex *dataws3,
				    int *wacoefStart
				    );




#ifdef __cplusplus
}
#endif




#endif
