#include "WA_cudaFFT.h"

bool ifft_CUDA_data(inputData *id)
{ 
  size_t rank = id->numberOfDimensions;
  size_t numElems = id->numberOfElements;
  //Device memory for data
  size_t mem_size = sizeof(cuDoubleComplex)*numElems;
  cuDoubleComplex* d_data;
  CudaSafeCall(cudaMalloc((void**)&d_data,mem_size));

  //Copy host data to device
  CudaSafeCall(cudaMemcpy(d_data, id->data, mem_size, cudaMemcpyHostToDevice));

  cufftHandle plan;
  switch(rank){
  case 1:
    CufftSafeCall(cufftPlan1d(&plan, numElems, CUFFT_Z2Z,1));
    break;
  case 2:
    CufftSafeCall(cufftPlan2d(&plan, id->sizeOfDimension[0],id->sizeOfDimension[1], CUFFT_Z2Z));
    break;
  case 3:
    CufftSafeCall(cufftPlan3d(&plan, id->sizeOfDimension[0],id->sizeOfDimension[1],id->sizeOfDimension[2], CUFFT_Z2Z));
    break;
  default:
    fprintf(stderr,"%zuD CUDA inverse FFT not supported at the moment!\n",rank); 
    return false;
  }

  CufftSafeCall(cufftExecZ2Z(plan, (cufftDoubleComplex *)d_data, (cufftDoubleComplex *)d_data, CUFFT_INVERSE));
  CufftSafeCall(cufftDestroy(plan));

  thrustMultiplyOnDevice(d_data,numElems,1./sqrt(numElems));

  //Copy device data to host
  CudaSafeCall(cudaMemcpy(id->data, d_data, mem_size,cudaMemcpyDeviceToHost));

  CudaSafeCall(cudaFree(d_data));

  return true; 
}


bool fft_CUDA_data(inputData *id, cuDoubleComplex* d_data)
{
  size_t rank = id->numberOfDimensions;
  size_t numElems = id->numberOfElements;

  /*
  //Device memory for data
  size_t mem_size = sizeof(cuDoubleComplex)*numElems;
  cuDoubleComplex* d_data;
  CudaSafeCall(cudaMalloc((void**)&d_data,mem_size));

  //Copy host data to device
  CudaSafeCall(cudaMemcpy(d_data, id->data, mem_size, cudaMemcpyHostToDevice));
  */

  cufftHandle plan;
  switch(rank){
  case 1:
    CufftSafeCall(cufftPlan1d(&plan, numElems, CUFFT_Z2Z,1));
    break;
  case 2:
    CufftSafeCall(cufftPlan2d(&plan, id->sizeOfDimension[0],id->sizeOfDimension[1], CUFFT_Z2Z));
    break;
  case 3:
    CufftSafeCall(cufftPlan3d(&plan, id->sizeOfDimension[0],id->sizeOfDimension[1],id->sizeOfDimension[2], CUFFT_Z2Z));
    break;
  default:
    fprintf(stderr,"%zuD CUDA inverse FFT not supported at the moment!\n",rank); 
    return false;
  }

  CufftSafeCall(cufftExecZ2Z(plan, (cufftDoubleComplex *)d_data, (cufftDoubleComplex *)d_data, CUFFT_FORWARD));
  CufftSafeCall(cufftDestroy(plan));

  thrustMultiplyOnDevice(d_data,numElems,1./sqrt(numElems));

  /*
  //Copy device data to host
  CudaSafeCall(cudaMemcpy(id->data, d_data, mem_size,cudaMemcpyDeviceToHost));
  CudaSafeCall(cudaFree(d_data));
  */

  return true; 
}
