#ifndef WA_CUDA_ERRORHANDLING_H_INCLUDED
#define WA_CUDA_ERRORHANDLING_H_INCLUDED

//Undef this to disable errorchecking
#define CUDA_ERROR_CHECK

#include <stdio.h>
#include <cuda.h>
#include <cuda_runtime_api.h>
#include <cufft.h>
#include <curand.h>

#ifdef __cplusplus
extern "C" {
#endif
  
#define CudaSafeCall( err ) __cudaSafeCall( err, __FILE__, __LINE__ )
#define CufftSafeCall( err ) __cufftSafeCall( err, __FILE__, __LINE__ )
#define CudarandSafeCall( err ) __cudarandSafeCall( err, __FILE__, __LINE__ )
#define CudaCheckError() __cudaCheckError( __FILE__, __LINE__ )
  
  inline void __cudaSafeCall( cudaError_t err, const char *file, const int line )
  {
#ifdef CUDA_ERROR_CHECK
    if ( cudaSuccess != err ){
      fprintf( stderr, "cudaSafeCall() failed at %s:%i : %s\n",
	       file, line, cudaGetErrorString( err ) );
      exit( -1 );
    }
#endif
    
  return;
  }
  
  inline void __cufftSafeCall( cufftResult err, const char *file, const int line )
  {
#ifdef CUDA_ERROR_CHECK
    if ( CUFFT_SUCCESS != err ){
      fprintf( stderr, "cufftSafeCall() failed at %s:%i\n", file, line);
      exit( -1 );
    }
#endif
    
    return;
  }

  inline void __cudarandSafeCall( curandStatus_t err, const char *file, const int line )
  {
#ifdef CUDA_ERROR_CHECK
    if ( CURAND_STATUS_SUCCESS != err )
      {
	fprintf( stderr, "cudarandSafeCall() failed at %s:%i\n", file, line);
	exit( -1 );
      }
#endif
    
    return;
  }


    
  inline void __cudaCheckError( const char *file, const int line )
  {
#ifdef CUDA_ERROR_CHECK
    cudaError_t err = cudaGetLastError();
    if ( cudaSuccess != err ){
      fprintf( stderr, "cudaCheckError() failed at %s:%i : %s\n",
	       file, line, cudaGetErrorString( err ) );
      exit( -1 );
    }
    /*
      err = cudaDeviceSynchronize();
      if( cudaSuccess != err ){
      fprintf( stderr, "cudaCheckError() with sync failed at %s:%i : %s\n",
      file, line, cudaGetErrorString( err ) );
      exit( -1 );
      }
    */
#endif
    
    return;
  }
  
#ifdef __cplusplus
}
#endif

#endif
