#include "WA_frequencypartition.h"


void PrintFrequencyPartition(frequencyPartition *freq_par)
{
  printf("Printing frequency partition:\n");
  for (size_t i = 0; i<= freq_par->maxJ; ++i){
    printf("j=%zu, length of m=%zu, start of m=%zu\n", i, freq_par->lengthOfM[i],freq_par->startOfM[i]);
  }
  printf("Done printing frequency partition.\n");
}

void FreeFrequencyPartition(frequencyPartition *freq_par)
{  
  free(freq_par->lengthOfM);
  free(freq_par->startOfM);
}

bool UniformFrequencyPartition(const size_t n, frequencyPartition *freq_par)
{



  if(n < 8){
    fprintf(stderr,"The number of elements in the smallest dimension must be at least 8! Currently it is %zu.\n", n);
    return false;
  }

  freq_par->maxJ = log2(n)/2;
  freq_par->lengthOfM = NULL;
  freq_par->startOfM = NULL;
  freq_par->lengthOfM = (size_t*)calloc((freq_par->maxJ+1),sizeof(size_t));
  freq_par->startOfM = (size_t*)calloc((freq_par->maxJ+1),sizeof(size_t));
  if(freq_par->lengthOfM == NULL || freq_par->startOfM == NULL){
    fprintf(stderr,"Error when trying to allocate memory for frequency partition!\n");
    FreeFrequencyPartition(freq_par);
    return false;
  }

  freq_par->lengthOfM[freq_par->maxJ] = n>>(freq_par->maxJ+1);
  return true;

}


bool OrthogonalFrequencyPartition(const size_t n, frequencyPartition *freq_par)
{
  if(n < 4){
    fprintf(stderr,"The number of elements in the smallest dimension must be at least 4! Currently it is %zu.\n", n);
    return false;
  }

  freq_par->lengthOfM = NULL;
  freq_par->startOfM = NULL;

  if(n == 4){
    freq_par->maxJ = 0;
    freq_par->lengthOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
    freq_par->startOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
    if(freq_par->lengthOfM == NULL || freq_par->startOfM == NULL){
      fprintf(stderr,"Error when trying to allocate memory for frequency partition!\n");
      FreeFrequencyPartition(freq_par);
      return false;
    }    
    freq_par->lengthOfM[0] = 2;

    freq_par->startOfM[0] = 0;
    
    return true;  

  } else if(n == 8){
    freq_par->maxJ = 1;
    freq_par->lengthOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
    freq_par->startOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
    if(freq_par->lengthOfM == NULL || freq_par->startOfM == NULL){
      fprintf(stderr,"Error when trying to allocate memory for frequency partition!\n");
      FreeFrequencyPartition(freq_par);
      return false;
    }    
    freq_par->lengthOfM[0] = 2;
    freq_par->lengthOfM[1] = 2;

    freq_par->startOfM[0] = 0;
    freq_par->startOfM[1] = 1;
    
    return true;
    
  } else if(n == 16){
    freq_par->maxJ = 1;
    freq_par->lengthOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
    freq_par->startOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
    if(freq_par->lengthOfM == NULL || freq_par->startOfM == NULL){
      fprintf(stderr,"Error when trying to allocate memory for frequency partition!\n");
      FreeFrequencyPartition(freq_par);
      return false;
    }    
    freq_par->lengthOfM[0] = 2;
    freq_par->lengthOfM[1] = 4;

    freq_par->startOfM[0] = 0;
    freq_par->startOfM[1] = 1;

    return true;

  } else if(n == 32){
    freq_par->maxJ = 2;
    freq_par->lengthOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
    freq_par->startOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
    if(freq_par->lengthOfM == NULL || freq_par->startOfM == NULL){
      fprintf(stderr,"Error when trying to allocate memory for frequency partition!\n");
      FreeFrequencyPartition(freq_par);
      return false;
    }    
    freq_par->lengthOfM[0] = 2;
    freq_par->lengthOfM[1] = 6;
    freq_par->lengthOfM[2] = 4;

    freq_par->startOfM[0] = 0;
    freq_par->startOfM[1] = 1;
    freq_par->startOfM[2] = 3;

    return true;
  } else if(n == 64){
    freq_par->maxJ = 2;
    freq_par->lengthOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
    freq_par->startOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
    if(freq_par->lengthOfM == NULL || freq_par->startOfM == NULL){
      fprintf(stderr,"Error when trying to allocate memory for frequency partition!\n");
      FreeFrequencyPartition(freq_par);
      return false;
    }    
    freq_par->lengthOfM[0] = 2;
    freq_par->lengthOfM[1] = 6;
    freq_par->lengthOfM[2] = 8;

    freq_par->startOfM[0] = 0;
    freq_par->startOfM[1] = 1;
    freq_par->startOfM[2] = 3;

    return true;
  }

  //Calculate how much memory is required
  //TODO: Could this be calculated analytically somehow?
  size_t r = 4; size_t f = 7; size_t s = 80;
  freq_par->maxJ = 2;
  while(s<n){
    f = 2*f-1;
    s = s+4*r*f;
    freq_par->maxJ++;
    r = 2*r;
  }
  
  //Allocate enough memory
  freq_par->lengthOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
  freq_par->startOfM = (size_t*)malloc((freq_par->maxJ+1)*sizeof(size_t));
  if(freq_par->lengthOfM == NULL || freq_par->startOfM == NULL){
    fprintf(stderr,"Error when trying to allocate memory for frequency partition!\n");
    FreeFrequencyPartition(freq_par);
    return false;
  }      
  
  //Generate partition
  freq_par->lengthOfM[0] = 2;
  freq_par->lengthOfM[1] = 6;
  freq_par->lengthOfM[2] = 10;
    
  freq_par->startOfM[0] = 0;
  freq_par->startOfM[1] = 1;
  freq_par->startOfM[2] = 3;

  r = 4; f = 7; s = 80; size_t olds = 80;
  for (size_t j = 3; j<=freq_par->maxJ; ++j){
    f = 2*f-1;
    s = s+4*r*f;
    
    freq_par->startOfM[j] = (1<<(j-1))+1;
    if(s<=n){
      freq_par->lengthOfM[j] = f+freq_par->startOfM[j];
    } else {
      freq_par->lengthOfM[j] = (n-olds)/(4.*r)+freq_par->startOfM[j];
    }    

    r = 2*r;
    olds = s;
  }

  return true;

}
