#ifndef WA_READDATA_H_INCLUDED
#define WA_READDATA_H_INCLUDED

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdbool.h>
#include <string.h>
#include <complex.h>
#include <math.h>

#include "WA_cuda.h"
#include "WA_cudaFFT.h"
#include "WA_datatypes.h"

/*
  Read double complex data from dir to id. (Including dimension header)
 */
bool readDoubleComplexData(const char* dir, inputData *id);

/*
  Writes double complex coefficients from id to dir. (Including dimension header)
 */
bool writeDoubleComplexDatafile(const char* dir, inputData *id);

/*
  Reads double complex coefficients (including dimension header) from dir and returns the read data. Allocates the size of id accordingly.
 */
double complex* readDoubleComplexCoefficientDatafile(const char* dir, inputData *id);

/*
  Reads dimension header information from fp to id.
 */
bool readDimensionHeader(inputData *id, FILE *fp);

/*
  Writes dimension header information from fp to id.
 */
bool writeDimensionHeader(const inputData *id, FILE *fp);

/*
  Basic inputData handling routines
 */
void initializeInputData(inputData *id);
void printInputData(const inputData *id, size_t nShowElements);
void freeInputData(inputData *id);

#endif
