%% WaveAtomC_Real matlab example 2
% In this example we plot 1D Wave Atom
% This example should plot the same Wave Atom as the WaveAtom-1.1.1
% test_wa1.m when N = 512, pat='q', tp='ortho', and (j,m) = (3,5)

%% Set the size of data
dims = 512;

%% Define used partition type
partition = 1; % 1 = multiscale 2 = monoscale

%% Define filenames
modifiedcoefficientfile = '../validation/modcoef.dat';
outputdatafile = '../validation/out.dat';

%% Define program locations
WA_fwa = '../WA_fwa'; % Forward WA transform
WA_iwa = '../WA_iwa'; % Inverse WA transform

%% Define datatypes used in input data
% intType should be the type corresponding to size_t c-type.
% Usually 'int32' for 32-bit system and 'int64' for 64-bit system.
% Currently WaveAtom package supports only dataType 'double' 
% It is straightforward to modify WaveAtom package code to handle 'float'
intType = 'int64';
dataType = 'double';

%% Define parameters concerning the use of FFTW
% See '../WA_testrun.sh' for explanations
savewisdom = 1;
readwisdom = 1;
fwawisdomfile = '../validation/fwawisdom.dat';
iwawisdomfile = '../validation/iwawisdom.dat';
fftwmeasure = 0;
fftwthreads = 1;
fftwtimelimit = -1;

%% Specify the scale, frequency position, and spatial translation
% The following function call can be used to deduce bounds
% for scale and frequency position.
[J_max, M_min, M_max] = WA_frequencyPartition(min(dims),partition)

% The following function call can be used to generate all
% valid scale and frequency position combinations
% WA_generateValidJMcombinations(dims,partition)

% Scale
% Valid scale is 0,...,J_max
jj = 2;

% Frequency position
% Valid frequency position in 1D is M_min,...,M_length-1
m = [4];

% Spatial translation
% Valid translation is between [1,...,1],...,(dims/min(dims))*2^(jj+1)
S_max = (dims/min(dims))*2^(jj+1)
s = 5;

% The following function call can be used to deduce whether a
% combination of scale and frequency position is valid
% WA_validJMcombination(jj,m,dims,partition)

% The following function call can be used to generate spatial
% translation indices corresponding to a given scale and
% frequency position combination
% WA_calculateBlockIndices(jj,m,dims,partition)

% The following function call can be used to calculate given
% WA coefficient position in the WA coefficient vector when the
% scale, frequency position, and spatial translation is given.
% WA_calculateCoefIndex(jj,m,s,dims,partition)

%% Generate coefficient vector where the above specified coefficient is one
coefs = zeros(prod(dims),1);
coefs(WA_calculateCoefIndex(jj,m,s,dims,partition)) = 1;
 
%% Write coefficients into a file
fprintf(1,'Writing coefficients into the file %s.\n',modifiedcoefficientfile)
tic();
WA_writeCoefficients(modifiedcoefficientfile,coefs,dims,intType,dataType);
toc()

%% Perform inverse WA transform
fprintf(1,'Calculating inverse WA transform.\n')
tic();
[is,iw] = unix([WA_iwa,...
    ' -coefdatafile ', modifiedcoefficientfile, ...
    ' -outputdatafile ', outputdatafile, ...
    ' -partition ', num2str(partition), ...
    ' -savewisdom ', num2str(savewisdom), ...
    ' -readwisdom ', num2str(readwisdom), ...
    ' -wisdomfile ', iwawisdomfile, ...
    ' -fftwmeasure ', num2str(fftwmeasure), ...
    ' -fftwthreads ', num2str(fftwthreads), ...
    ' -fftwtimelimit ', num2str(fftwtimelimit), ...
    ]);
if(is~=0)
    iw
    error('Backward Wave Atom tMuttransform failed.')
end
toc()

%% Read result
fprintf(1,'Reading resulting output.\n')
tic();
output =  WA_readDatainput(outputdatafile,intType,dataType);
toc()

%% Plot WaveAtom
figure; plot(output);  title(sprintf('j=%d, m=%d, s=%d, spatial domain',jj,m,s));
figure; plot(abs(fftshift(fft2(output))));  title(sprintf('j=%d, m=%d, s=%d, frequency domain',jj,m,s));
