function [] = WA_writeCoefficients(fileName, coefficients, dimInfo,intType, dataType)
% Writes the complex valued array coefficients consisting the Wave Atom 
% coefficients into file fileName.
% dimInfo is a vector consisting the size of dimensions.
% intType should be the type corresponding to size_t c-type.
% Usually 'int32' for 32-bit system and 'int64' for 64-bit system.
% Currently WaveAtom package supports only dataType 'double' 

%% Check that number of coefficients mathces given dimension info
    if (numel(coefficients) ~= prod(dimInfo))
        error('Number of coefficients does not match dimInfo!')
    end

%% Open output file
    [f,msg] = fopen(fileName,'w');
    if(isempty(msg)==0)
        error(msg)
        return
    end
    
%% Write dimension header
    c = fwrite(f,length(dimInfo),intType);
    if (c ~= 1)    
        fclose(f);
        error('Failed to write the number of dimensions.')
    end 

    c = fwrite(f,dimInfo,intType);
    if (c ~= length(dimInfo))
        fclose(f);
        error('Failed to write the size of dimensions.')
    end    

%% Write coefficients into output file
    temp = coefficients(:).';    
    c = fwrite(f, temp, dataType);
    if (c ~= numel(coefficients))    
        fclose(f);
        error('Failed to write the correct number of coefficients.')
    end    

%% Close output file
    fclose(f);
    return

end
