#include "WA_readdata.h"

bool ifftND_data(inputData *id,unsigned flags)
{  
  int rank = (int)id->numberOfDimensions;  
  int *n = malloc(rank*sizeof(int));
  if(n==NULL){
    fprintf(stderr,"Failed to calculate inverse FFT.\n");
    return false;
  }
  for(int i = 0; i<rank; ++i){
    n[i] = (int)id->sizeOfDimension[i];
  }
  
  fftw_complex *in  = (fftw_complex*) id->data;
  double *out = (double*) in;

  fftw_plan p = fftw_plan_dft_c2r(rank, n, in, out, flags);
  fftw_execute(p);  
  fftw_destroy_plan(p);
  free(n);

  double normalization = 1./sqrt(id->numberOfElements);
  size_t coefsPos = 0;
  size_t coefsPosShift = 0;
  size_t rowElems = id->numberOfElements/id->sizeOfDimension[rank-1];
  for (size_t d1 = 0; d1<rowElems;++d1){	  
    for(size_t d2 = 0; d2<id->sizeOfDimension[rank-1];++d2){
      out[coefsPos++] = out[coefsPosShift++]*normalization;
    }
    coefsPosShift += 2;
  } 

  return true;
}

bool fftND_data(inputData *id,unsigned flags)
{
  int rank = (int)id->numberOfDimensions;
  int *n = malloc(rank*sizeof(int));
  if(n==NULL){
    fprintf(stderr,"Failed to calculate FFT.\n");
    return false;
  }

  for(int i = 0; i<rank; ++i){
    n[i] = id->sizeOfDimension[i];
  }
  
  double *in  = id->data;
  fftw_complex *out = (fftw_complex*) in;
  fftw_plan p = fftw_plan_dft_r2c(rank, n, in, out, flags);
  fftw_execute(p);  
  fftw_destroy_plan(p);
  free(n);
  
  double normalization = 1/sqrt(id->numberOfElements);
  for(size_t i = 0; i<id->allocatedNumberOfElements/2;++i){
    out[i] *= normalization;
  } 
  

  return true;
}

bool writeDoubleRealDatafile(const char* dir, inputData *id)
{
  FILE *fp = fopen(dir,"wb");
  if(fp==NULL){
    fprintf(stderr,"Failed to open file %s for writing.\n",dir);
    return false;
  }  

  if(!writeDimensionHeader(id, fp)){
    fprintf(stderr,"Failed to write data dimensions.\n");
    return false;
  }
  
  if (fwrite(id->data,sizeof(double),id->numberOfElements,fp) != id->numberOfElements){
    fprintf(stderr,"Failed to write output data.\n");
    return false;
  }
  
  fclose(fp);
  return true;
  
}

double* readDoubleRealCoefficientDatafile(const char* dir, inputData *id)
{
  initializeInputData(id);
  FILE *fp = fopen (dir , "rb" );
  if (fp==NULL) {
    fprintf(stderr,"Could not open file %s.\n",dir);
    return false;
  }

  if(!readDimensionHeader(id,fp)){
    fprintf(stderr,"Failed to read the dimension header of %s\n",dir);
    fclose(fp);
    return false;
  }

  size_t rank = id->numberOfDimensions;
  
  id->numberOfElements = id->sizeOfDimension[rank-1];
  id->allocatedNumberOfElements = id->numberOfElements/2+1;
  
  for(size_t i=0;i<rank-1;i++){
    id->allocatedNumberOfElements *= id->sizeOfDimension[i];
    id->numberOfElements *= id->sizeOfDimension[i];    
  }
  
  id->data = (double*)fftw_malloc(id->allocatedNumberOfElements*sizeof(fftw_complex));
    
  if (id->data == NULL){
    fprintf(stderr,"Failed to allocate memory for coefficients.\n");
    fclose(fp);
    freeInputData(id);    
    return NULL;
  }
  
  // allocate memory to containt the coefficients and enough extra space to do FFTs
  double *data = (double*) fftw_malloc(id->numberOfElements*sizeof(double));
  if (data == NULL) {
    fprintf(stderr,"Could not allocate enough memory for data!");
    freeInputData(id);
    fclose (fp);
    return NULL;
  }

  // copy the file into the buffer:
  if (fread(data,sizeof(double),id->numberOfElements,fp) != id->numberOfElements) {
    fprintf(stderr,"Error wile reading data from file %s.\n",dir);
    fftw_free(data);
    freeInputData(id);
    fclose(fp);
    return NULL;
  }
    
  fclose (fp);

  return data;  
}

double* readDoubleRealCoefficientDatafileFFTWform(const char* dir, inputData *id, bool uniformpartition)
{
  initializeInputData(id);
  FILE *fp = fopen (dir , "rb" );
  if (fp==NULL) {
    fprintf(stderr,"Could not open file %s.\n",dir);
    return false;
  }

  if(!readDimensionHeader(id,fp)){
    fprintf(stderr,"Failed to read the dimension header of %s\n",dir);
    fclose(fp);
    return false;
  }

  size_t rank = id->numberOfDimensions;
  
  id->numberOfElements = id->sizeOfDimension[rank-1];
  id->allocatedNumberOfElements = id->numberOfElements/2+1;
  
  for(size_t i=0;i<rank-1;i++){
    id->allocatedNumberOfElements *= id->sizeOfDimension[i];
    id->numberOfElements *= id->sizeOfDimension[i];    
  }
  
  id->data = (double*)fftw_malloc(id->allocatedNumberOfElements*sizeof(fftw_complex));
    
  if (id->data == NULL){
    fprintf(stderr,"Failed to allocate memory for coefficients.\n");
    fclose(fp);
    freeInputData(id);    
    return NULL;
  }
  
  // allocate memory to containt the coefficients and enough extra space to do FFTs
  size_t sizeOfLastDimension = id->sizeOfDimension[rank-1];
  size_t sizeOfSmallestDimension = id->sizeOfDimension[0];
  for(size_t i = 1; i<rank;++i){
    if (id->sizeOfDimension[i]<sizeOfSmallestDimension){
      sizeOfSmallestDimension=id->sizeOfDimension[i];
    }
  }

  size_t ratiosum = id->sizeOfDimension[0]/sizeOfSmallestDimension;
  size_t ratioprod = ratiosum; 
  for(size_t i = 1; i<rank;++i){
    size_t ratio = id->sizeOfDimension[i]/sizeOfSmallestDimension;
    ratiosum += ratio;
    ratioprod *= ratio;
  }

  frequencyPartition freq_par;
  bool (*partitiofunc)(const size_t, frequencyPartition *);
  if (uniformpartition){
    partitiofunc = UniformFrequencyPartition;
  } else {
    partitiofunc = OrthogonalFrequencyPartition;
  }
  
  if (!partitiofunc(sizeOfSmallestDimension, &freq_par)){
    return false;
  }
  PrintFrequencyPartition(&freq_par);  

  //Calculate required memory
  size_t requiredMemForCoef = 0;
  for(size_t j = 0; j<=freq_par.maxJ;++j){
    size_t validMCombinations = pow(freq_par.lengthOfM[j],rank) - pow(freq_par.startOfM[j],rank);
    requiredMemForCoef += 2*validMCombinations*
      (1<<((j+1)*(rank-1)))*
      ratioprod*
      sizeOfSmallestDimension/sizeOfLastDimension* 
      (((1<<j)*
	sizeOfLastDimension)/sizeOfSmallestDimension+1);
  }



  double *data = (double*) fftw_malloc(requiredMemForCoef*sizeof(double));
  if (data == NULL) {
    fprintf(stderr,"Could not allocate enough memory for data!");
    freeInputData(id);
    fclose (fp);
    return NULL;
  }

  // copy the file into the buffer:

  double *datapos = data;
  
  for(size_t j = 0; j<=freq_par.maxJ;++j){
    size_t validMCombinations = pow(freq_par.lengthOfM[j],rank) - pow(freq_par.startOfM[j],rank);
    size_t rowelemsj = (1<<((j+1)*(rank-1)))*ratioprod*sizeOfSmallestDimension/sizeOfLastDimension;
    size_t ratioNcoef = sizeOfLastDimension/sizeOfSmallestDimension*(1<<(j+1));         

    for(size_t m = 0; m< validMCombinations; ++m){   
      for(size_t d = 0; d<rowelemsj; ++d){
	if (fread(datapos,sizeof(double),ratioNcoef,fp) != ratioNcoef) {
	  fprintf(stderr,"Error wile reading data from file %s.\n",dir);
	  fftw_free(data);
	  freeInputData(id);
	  fclose(fp);
	  return NULL;
	}
	datapos += ratioNcoef+2;
      }
    }
  }
  
  fclose (fp);

  return data;  
}


bool readDoubleRealDataFFTWform(const char* dir, inputData *id)
{
  printf("Reading inputdatafile '%s'...\n",dir);
  
  initializeInputData(id);

  FILE *fp = fopen(dir,"rb");
  if(fp==NULL){
    fprintf(stderr,"Failed to open file '%s'.\n",dir);
    return false;
  }

  if(!readDimensionHeader(id,fp)){
    fprintf(stderr,"Failed to read the dimension header of %s\n",dir);
    fclose(fp);
    return false;
  }

  id->numberOfElements = id->sizeOfDimension[id->numberOfDimensions-1];;
  id->allocatedNumberOfElements = id->numberOfElements+2;
  
  for(size_t i=0;i<id->numberOfDimensions-1;i++){
    id->allocatedNumberOfElements *= id->sizeOfDimension[i];
    id->numberOfElements *= id->sizeOfDimension[i];    
  }
  
  id->data = (double*)fftw_malloc(id->allocatedNumberOfElements*sizeof(double));
  if (id->data == NULL){
    fprintf(stderr,"Failed to allocate memory for %zu elements of doubles.\n", id->allocatedNumberOfElements);
    fclose(fp);
    freeInputData(id);    
    return false;
  }  


  size_t el = id->sizeOfDimension[id->numberOfDimensions-1];  
  size_t nl = id->numberOfElements/el;
  for(size_t i=0;i<nl;i++){
    if(fread(id->data+i*(el+2),sizeof(double),el,fp)!=el){
      fprintf(stderr,"Failed to read dataelements\n");
      fclose(fp);
      freeInputData(id);
      return false;
    }
  }
  
  fclose(fp);
  printf("Reading inputdatafile done.\n");
  
  return true;  
}


void printInputData(const inputData *id, size_t nShowElements)
{
  printf("Number of dimensions: %zu\n", id->numberOfDimensions);

  if(id->sizeOfDimension == NULL){
    return;
  }  

  for(size_t i = 0; i < id->numberOfDimensions; ++i){
    printf("\tDimension %zu: %zu\n", i,id->sizeOfDimension[i]);    
  }  

  printf("Number of physical elements: %zu\n",id->numberOfElements);
  printf("Number of allocated elements: %zu\n",id->allocatedNumberOfElements);

  if (nShowElements > id->allocatedNumberOfElements) nShowElements = id->allocatedNumberOfElements;

  for(size_t i = 0; i < nShowElements; ++i){
    printf("%zu: %f\n", i, id->data[i]);    
  }
  printf("\n");

  return;
}

void initializeInputData(inputData *id)
{
  id->numberOfDimensions = 0;
  id->numberOfElements = 0;
  id->allocatedNumberOfElements = 0;
  id->sizeOfDimension = NULL;
  id->data = NULL;
}

void freeInputData(inputData *id)
{  
  if(id->sizeOfDimension != NULL){
    free(id->sizeOfDimension);
    id->sizeOfDimension = NULL;
  }

  if(id->data != NULL){
    fftw_free(id->data);
    id->data = NULL;
  }

  return;
}

bool readDimensionHeader(inputData *id, FILE *fp)
{
  if(fread(&id->numberOfDimensions,sizeof(size_t),1,fp)!=1){
    fprintf(stderr,"Failed to read the number of dimensions.\n");
    return false;
  }
  
  id->sizeOfDimension = (size_t*)malloc(id->numberOfDimensions*sizeof(size_t));
  if (id->sizeOfDimension == NULL){
    fprintf(stderr,"Failed to allocate memory for sizeOfDimensions.\n");
    return false;
  }
  
  if(fread(id->sizeOfDimension,sizeof(size_t),id->numberOfDimensions,fp) != id->numberOfDimensions){
    fprintf(stderr,"Failed to read the number of elements in dimensions.\n");
    freeInputData(id);
    return false;
  }
  return true;
}
  

bool writeDimensionHeader(const inputData *id, FILE *fp)
{
  
  if (fwrite(&(id->numberOfDimensions),sizeof(size_t),1,fp) != 1){
    return false;
  }

  if(fwrite(id->sizeOfDimension,sizeof(size_t),id->numberOfDimensions,fp)!=id->numberOfDimensions){
    return false;
  }

  return true;
}
