#ifndef WA_READDATA_H_INCLUDED
#define WA_READDATA_H_INCLUDED

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdbool.h>
#include <string.h>
#include <complex.h>
#include <math.h>
#include <fftw3.h>
#include "WA_frequencypartition.h"
#include "WA_datatypes.h"

/*
  Read double real data from dir to id in FFTW form. (Including dimension header)
 */
bool readDoubleRealDataFFTWform(const char* dir, inputData *id);

/*
  Writes double real coefficients from id to dir. (Including dimension header)
 */
bool writeDoubleRealDatafile(const char* dir, inputData *id);

/*
  Reads double real coefficients (including dimension header) in normal or in FFTW form from dir and returns the read data. Allocates the size of id accordingly.
 */
double* readDoubleRealCoefficientDatafile(const char* dir, inputData *id);
double* readDoubleRealCoefficientDatafileFFTWform(const char* dir, inputData *id, bool uniformpartition);

/*
  Reads dimension header information from fp to id.
 */
bool readDimensionHeader(inputData *id, FILE *fp);

/*
  Writes dimension header information from fp to id.
 */
bool writeDimensionHeader(const inputData *id, FILE *fp);

/*
  Basic inputData handling routines
 */
void initializeInputData(inputData *id);
void printInputData(const inputData *id, size_t nShowElements);
void freeInputData(inputData *id);

/*
  Perform FFT or inverse FFT to input data using given FFTW flags.
 */
bool fftND_data(inputData *id,unsigned flags);
bool ifftND_data(inputData *id, unsigned flags);

#endif
