#include "WA_fwa.h"

bool fwa2(const inputData *id, double** coefficients, bool uniformpartition,unsigned fftw_flags)
{
  /*
    Check that input data has correct number of dimensions.
   */

  printf("Calculating 2D forward wave atom transform...\n");

  if(id->numberOfDimensions != 2){
    printf("Number of dimensions is not two! Instead, it is (%zu).\n", id->numberOfDimensions);
    return false;
  }

  /* 
     Reserve enough memory for different arrays, perform other initializations, and generate frequency partition.
   */

  size_t sizeOfLastDimension = id->sizeOfDimension[1];
  size_t sizeOfSmallestDimension = id->sizeOfDimension[0];
  for(size_t i = 1; i<2;++i){
    if (id->sizeOfDimension[i]<sizeOfSmallestDimension){
      sizeOfSmallestDimension=id->sizeOfDimension[i];
    }
  }
  
  size_t ratiosum = id->sizeOfDimension[0]/sizeOfSmallestDimension;
  size_t ratioprod = ratiosum; 
  for(size_t i = 1; i<2;++i){
    size_t ratio = id->sizeOfDimension[i]/sizeOfSmallestDimension;
    ratiosum += ratio;
    ratioprod *= ratio;
  }
  
  frequencyPartition freq_par;
  bool (*partitiofunc)(const size_t, frequencyPartition *);
  if (uniformpartition){
    partitiofunc = UniformFrequencyPartition;
  } else {
    partitiofunc = OrthogonalFrequencyPartition;
  }
  
  if (!partitiofunc(sizeOfSmallestDimension, &freq_par)){
    return false;
  }
  
  PrintFrequencyPartition(&freq_par);
  
  size_t maximumLengthOfM = freq_par.lengthOfM[0];
  for(size_t i = 1; i<=freq_par.maxJ;++i){
    if(maximumLengthOfM < freq_par.lengthOfM[i]){
      maximumLengthOfM = freq_par.lengthOfM[i];
    }      
  }
  
  size_t memforcoefs = 0;
  if(freq_par.maxJ>0){
    for (size_t j=0; j<=freq_par.maxJ-1;++j){
      memforcoefs += (pow(freq_par.lengthOfM[j],2)-pow(freq_par.startOfM[j],2))*ratioprod*(1<<(2*(j+1)));
    }
  }
  memforcoefs += 
    (pow(freq_par.lengthOfM[freq_par.maxJ],2)-pow(freq_par.startOfM[freq_par.maxJ],2))*
    (1<<((freq_par.maxJ+1)))*ratioprod*
    sizeOfSmallestDimension/sizeOfLastDimension*
    (sizeOfLastDimension/sizeOfSmallestDimension*(1<<(freq_par.maxJ+1))+2);
  
  fftw_complex* coeff = fftw_malloc(memforcoefs*sizeof(fftw_complex));
  if (coeff == NULL){
    fprintf(stderr,"Failed to reserve memory for the coefficients.\n");
    return false;
  }
  
  fftw_complex* wacoef = (fftw_complex*) malloc(ratiosum*maximumLengthOfM*(2<<freq_par.maxJ)*sizeof(fftw_complex));  
  fftw_complex* wacoefConj = (fftw_complex*) malloc(ratiosum*maximumLengthOfM*(2<<freq_par.maxJ)*sizeof(fftw_complex));  
  long long int* wacoefStart = (long long int*) malloc(2*maximumLengthOfM*sizeof(long long int));  
  long long int *sijoituskohtaModJ = (long long int*)malloc(2*sizeof(long long int));
  long long int *sijoituskohtaModDim = (long long int*)malloc(2*sizeof(long long int));

  int *nFftwf = malloc(2*sizeof(int));

  size_t *ratioNcoef = malloc(2*sizeof(size_t));
 
  if (wacoef == NULL || wacoefConj == NULL || wacoefStart == NULL || 
      sijoituskohtaModJ == NULL || sijoituskohtaModDim == NULL ||
      nFftwf == NULL || ratioNcoef == NULL){
    printf("Error when reserving memory.");
    FreeFrequencyPartition(&freq_par);
    free(wacoef);
    free(wacoefConj);
    free(wacoefStart);
    free(sijoituskohtaModJ);
    free(sijoituskohtaModDim);
    free(nFftwf);
    free(ratioNcoef);
    fftw_free(coeff);
    return false;
  }    

  size_t pos;
  int posStart;
  
  for(size_t h = 0; h<2;++h){
    ratioNcoef[h] = id->sizeOfDimension[h]/sizeOfSmallestDimension;
  }

  fftw_complex* tempmemory;
  fftw_complex* tempmemorystart = (fftw_complex*)coeff;

  /*
    Loop over scale j
        Calculate required Mayer's Wavelet values for scale j.
        Loop over frequency partition m=(m1,m2)
	    Calculate coefficient set for fixed j and m by
            looping over right/left bumps i=(i1,i2) and by
	    looping over 2D frequency area specified by (j,m) using index s=(s1,s2).
	End
	Perform FFT to each calculated coefficient set j and m.
	Normalize each coefficient set to get final coefficients corresponding to j and m.
    End
    Return calculated coefficents.
   */

  for (size_t j=0; j<=freq_par.maxJ;++j){

    printf("Scale j=%zu...\n",j);

    size_t lm = freq_par.lengthOfM[j];
    size_t sm = freq_par.startOfM[j];
    
    for(size_t h = 0; h<2;++h){
      ratioNcoef[h] <<=1;
    }      

    if(lm==0) continue;
    tempmemory = tempmemorystart;
    size_t nelemsj = (1<<((j+1)*2))*ratioprod;
    size_t rowelemsj = (1<<(j+1))*ratioprod*sizeOfSmallestDimension/sizeOfLastDimension;
    double b = 1<<j;

    for(size_t i = 0; i<2; ++i){
      nFftwf[i] =  (int)ratioNcoef[i];
    }   

    pos = 0;
    posStart = 0;

    for (size_t curDim = 0; curDim <2;++curDim){	
      bool wacoefsCalculatedAlready = false;            
      size_t copypos = 0;
      for (size_t d = 0; d<curDim;++d){
	if(id->sizeOfDimension[d] == id->sizeOfDimension[curDim]){
	  wacoefsCalculatedAlready = true;
	  memcpy(&wacoefStart[posStart],&wacoefStart[d*lm],sizeof(long long int)*lm);
	  memcpy(&wacoef[pos],&wacoef[copypos],sizeof(fftw_complex)*ratioNcoef[d]*lm);
	  memcpy(&wacoefConj[pos],&wacoefConj[copypos],sizeof(fftw_complex)*ratioNcoef[d]*lm);
	  pos += ratioNcoef[d]*lm;
	  posStart += lm;
	  break;
	}
	copypos += ratioNcoef[d]*lm;
      }
     
      if(wacoefsCalculatedAlready == false){      
	double d,t;
	double complex c;
	for (size_t m = 0; m<freq_par.lengthOfM[j]/2;m++){
	  double fm = id->sizeOfDimension[curDim]/sizeOfSmallestDimension;
	  double dx = 1./(fm*b);

	  fm *= (2*m-2/3.)*b;
	  wacoefStart[posStart] = ceil(fm);
	  d = wacoefStart[posStart++]*dx-2*m;
	  t = M_PI_2*(d-0.5);

	  size_t tpos = pos;
	  double complex temp;

	  for(size_t i = 0; i < ratioNcoef[curDim]; ++i){
	    c = cos(t)-sin(t)*I;
	    if(4/3. < d || d < -2/3.) temp = 0;
	    else if (2/3. < d) temp=c*sin(M_PI_4*(1+cos(M_PI*(1-1.5*d))));
	    else temp= c*cos(M_PI_4*(1-cos(M_PI*(0.5-0.75*d))));
	    wacoef[ratioNcoef[curDim]-1+tpos-i] = temp;
	    wacoefConj[pos++] = conj(temp);
	    d += dx;  
	    t += M_PI_2*dx;
	  }	

	  fm += 4/3.*b*id->sizeOfDimension[curDim]/sizeOfSmallestDimension;
	  wacoefStart[posStart] = ceil(fm);
	  d = wacoefStart[posStart++]*dx-2*m-1;
	  t = M_PI_2*(d-0.5);
	  tpos = pos;
	  for(size_t i = 0; i < ratioNcoef[curDim]; ++i){
	    c = cos(t)-sin(t)*I;
	    if (d < -1/3. || 5/3. < d) temp = 0;
	    if (d < 1/3.) temp = c*sin(M_PI_4*(1+cos(M_PI*(1.5*d-0.5))));
	    else temp = c*cos(M_PI_4*(1-cos(M_PI*(0.75*d-0.25))));
	    wacoef[ratioNcoef[curDim]-1+tpos-i] = temp;
	    wacoefConj[pos++] = conj(temp);
	    d += dx;  
	    t += M_PI_2*dx;
	  }		
	}
      }
    }

    for(size_t m1 = 0; m1<lm;++m1){
      for(size_t m2 = 0; m2<lm;++m2){
	if(m1<sm && m2<sm){
	  continue;
	}
	
	memset(tempmemory,0,rowelemsj*(ratioNcoef[1]/2+1)*sizeof(fftw_complex));

	long long int temp = wacoefStart[m1];
	long long int sijoituskohtaModJ_orig = temp&(ratioNcoef[0]-1);
	long long int sijoituskohtaModDim_orig = temp&(id->sizeOfDimension[0]-1);

	size_t ks1 = m1*ratioNcoef[0];
	size_t ks2 = lm*ratioNcoef[0] + m2*ratioNcoef[1];

	fftw_complex* dataws1 = wacoefConj;		
	for(size_t i1 = 0; i1<2; ++i1){
	  if (sijoituskohtaModJ_orig < 0)
	    sijoituskohtaModJ_orig += ratioNcoef[0];
	  if (sijoituskohtaModDim_orig < 0)
	    sijoituskohtaModDim_orig += id->sizeOfDimension[0];

	  temp = wacoefStart[lm+m2];	      
	  sijoituskohtaModJ[1] = temp&(ratioNcoef[1]-1);
	  sijoituskohtaModDim[1] = temp&(id->sizeOfDimension[1]-1);

	  fftw_complex* dataws2 = wacoefConj;	  

	  for(size_t i2 = 0; i2<2; ++i2){	    
	    sijoituskohtaModJ[0] = sijoituskohtaModJ_orig;
	    sijoituskohtaModDim[0] = sijoituskohtaModDim_orig;

	    if (sijoituskohtaModJ[1] < 0)
	      sijoituskohtaModJ[1] += ratioNcoef[1];
	    if (sijoituskohtaModDim[1] < 0)
	      sijoituskohtaModDim[1] += id->sizeOfDimension[1];

	    for(size_t s1 = 0; s1<ratioNcoef[0];++s1){     
	      for(size_t s2 = 0; s2<ratioNcoef[1];++s2){
		if ((size_t)sijoituskohtaModJ[1]<ratioNcoef[1]/2+1){
		  fftw_complex kerroin = 
		    dataws1[ks1]*
		    dataws2[ks2];	 
		  size_t sijoituspossi =
		    sijoituskohtaModJ[0]*(ratioNcoef[1]/2+1)+
		    sijoituskohtaModJ[1];

		  fftw_complex datavalue;
		  size_t datapossi = 0;		  
		  if((size_t)sijoituskohtaModDim[1] > sizeOfLastDimension/2){    
		    if(sijoituskohtaModDim[0] != 0){
		      datapossi = id->sizeOfDimension[0]-sijoituskohtaModDim[0];
		    }
		    datapossi = datapossi*(sizeOfLastDimension/2+1)+
		      sizeOfLastDimension-sijoituskohtaModDim[1];
		    datavalue = conj(((fftw_complex*)id->data)[datapossi]);

		  } else {
		    datapossi = sijoituskohtaModDim[0]*(sizeOfLastDimension/2+1)+sijoituskohtaModDim[1];
		    datavalue = ((fftw_complex*)id->data)[datapossi];
		  }		  

		  tempmemory[sijoituspossi] += kerroin*datavalue;


		}

		sijoituskohtaModJ[1]++;
		sijoituskohtaModDim[1]++;

		if ((size_t)sijoituskohtaModJ[1] == ratioNcoef[1])
		  sijoituskohtaModJ[1] = 0;
		if ((size_t)sijoituskohtaModDim[1] == id->sizeOfDimension[1])
		  sijoituskohtaModDim[1] = 0;
		
		ks2++;

	      }
	      
	      if ((size_t)sijoituskohtaModDim[1] < ratioNcoef[1])
		sijoituskohtaModDim[1] += id->sizeOfDimension[1];
	      sijoituskohtaModDim[1] -= ratioNcoef[1];
	      sijoituskohtaModDim[1] &= id->sizeOfDimension[1]-1;


	      sijoituskohtaModJ[0]++;
	      sijoituskohtaModDim[0]++;
	      if ((size_t)sijoituskohtaModJ[0] == ratioNcoef[0])
		sijoituskohtaModJ[0] = 0;
	      if ((size_t)sijoituskohtaModDim[0] == id->sizeOfDimension[0])
		sijoituskohtaModDim[0] = 0;
	      
	      ks1++;
	      ks2 -= ratioNcoef[1];

	    }
	    temp = wacoefStart[lm+m2]+ratioNcoef[1]-1;	      
	    sijoituskohtaModJ[1] = (-temp)&(ratioNcoef[1]-1);
	    sijoituskohtaModDim[1] = (-temp)&(id->sizeOfDimension[1]-1);

	    dataws2 = wacoef;
	    ks1 -= ratioNcoef[0];
	  }
	  temp = wacoefStart[m1]+ratioNcoef[0]-1;	      
	  sijoituskohtaModJ_orig = (-temp)&(ratioNcoef[0]-1);
	  sijoituskohtaModDim_orig = (-temp)&(id->sizeOfDimension[0]-1);

	  dataws1 = wacoef;
	}
	tempmemory += rowelemsj*(ratioNcoef[1]/2+1);
      }
    }

    int validMCombinations = pow(lm,2)-pow(sm,2);    

    fftw_plan p = fftw_plan_many_dft_c2r(2,nFftwf,validMCombinations,
		  tempmemorystart,
		  NULL,1, rowelemsj*(ratioNcoef[1]/2+1),
		  (double*)tempmemorystart,
		  NULL,1,rowelemsj*(ratioNcoef[1]+2),fftw_flags);    
    fftw_execute(p);
    fftw_destroy_plan(p);


    size_t coefsPosShift = 0;
    double* tmstart = (double*) tempmemorystart;
    double* tm = (double*) tempmemorystart;
    
    double normalization =1./sqrt(nelemsj);
    for(int m = 0; m<validMCombinations; ++m){
      for (size_t d1 = 0; d1<rowelemsj;++d1){	  
	for(size_t d2 = 0; d2<ratioNcoef[1];++d2){
	  (*tmstart) = tm[coefsPosShift++]*normalization;
	  tmstart++;
	}
	coefsPosShift += 2;
      }
    }    
    tempmemorystart = (fftw_complex*) tmstart;
  }
  free(wacoef);    
  free(wacoefConj);    
  free(wacoefStart);
  free(sijoituskohtaModJ);
  free(sijoituskohtaModDim);
  free(nFftwf);
  FreeFrequencyPartition(&freq_par);
  free(ratioNcoef);

  (*coefficients) = (double*)coeff;

  return true;
}
